
//		Fortune.cpp - Random line plugin for Take Command


#define UNICODE					1
#define _UNICODE				1
#define _CRT_RAND_S

#include <stdio.h>
#include <tchar.h>
#include <math.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"

#include "codepages.cpp"


#define VER_MAJOR				1
#define VER_MINOR				2
#define VER_BUILD				4
#define VER_PATCH				2

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define PLUGIN_NAME				L"Fortune"


#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024

#define MAX_ARGS				1024

LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";


int ShellVerMajor = 0, ShellVerMinor = 0, ShellVerBuild = 0;

int RandomCount = 0;

wchar_t LastFortune[MAX_ARG_LEN] = L"";
int LastLine = -1;



#define BUFFER_SIZE				16384
#define MAX_LINES				131072
#define MAX_PRIORS				256
#define DEFAULT_PRIORS			10
#define MAX_MRU_ENTRIES			16
#define MAX_REG_VALUE_NAME_LEN	16384
#define MAX_FILENAME_LEN		16384

#define LF						0x000a
#define CR						0x000d
#define PILCROW					0x00b6
#define SECTION_MARK			0x00a7
#define QUOTE					0x0022
#define APOS					0x0027
#define NBSP					0x00a0
#define EMDASH					0x2014
#define LSQUOTE					0x2018
#define RSQUOTE					0x2019
#define LQUOTE					0x201c
#define RQUOTE					0x201d
#define BOM						0xfeff

#define MAX_TAB_SIZE			20
#define MAX_COLUMNS				256
#define MIN_COLUMNS				40

#define FILE_OEM				0
#define FILE_UTF16				1
#define FILE_UTF8				2
#define INVALID_CHAR			'?'

#define FLAG_BIASED				0x0001
#define FLAG_NO_ATTRIBS			0x0080
#define FLAG_FIX_QUOTES			0x0100
#define FLAG_RETURN_EOLS		0x0200



const wchar_t PluginFeatures[] = L"fortune,@fortune,_lastfline,_lastfortune";

const wchar_t PluginKeyName[] = L"SOFTWARE\\JPPlugins\\Fortune";
const wchar_t MRUValueName[] = L"Files MRU";

const wchar_t EnvVarName[] = L"FORTUNE";

const wchar_t HelpWindowTitle[] = L"Fortune Plugin Help";

const wchar_t FortunePurposeText[] = L"Display a random line from a text file.\n\n";

const wchar_t FortuneUsageText[] = L"FORTUNE /B /H /N:n /NB /Q /T:size /W:width filename\r\n\n\
\t/B\tBiased (earlier lines are more likely)\n\
\t/H\tHide attributions\n\
\t/N:\tNumber of previous fortunes to remember (10)  \n\
\t/NB\tno Byte Order Mark\n\
\t/Q\treplace Quote marks and apostrophes\n\
\t/T:\twidth of Tab stops (4)\n\
\t/W:\tdesired columns for Word-wrap\n\n\
If no filename is specified, this command will look for an\n\
environment variable named FORTUNE.\n";


const wchar_t f_FortuneUsageText[] = L"Return a random line from a text file.\n\n\
%@FORTUNE[filename,n,flags]\r\n\n\
\tfilename\tfile to read (required)\n\
\tn\tnumber of previous fortunes to remember (10)  \n\
\tflags\t1 = biased (earlier lines are more likely)\n\
\t\t128 = hide attributions\n";


const wchar_t _LastFLineUsageText[] = L"Returns the last line number from FORTUNE or @FORTUNE.  \n\n\
The line number is zero-based.  If neither FORTUNE nor\n\
@FORTUNE has returned a line yet, returns -1.\n";

const wchar_t _LastFortuneUsageText[] = L"Repeats the last line displayed by the FORTUNE command.\n\n\
The entire line is returned as is; no word-wrapping, tab\n\
expansion, or other character substitution is performed.  If\n\
FORTUNE has not yet displayed a line, returns an empty string.  \n";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



int GetShellVersion()
{
	wchar_t temp[SHORT_BUF_LEN] = L"%_4ver";
	wchar_t *sep1 = NULL;

	ExpandVariables( temp, 1 );
	ShellVerMajor = wcstol( temp, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellVerMinor = wcstol( sep1, NULL, 10 );
	}

	wcscpy_s( temp, SHORT_BUF_LEN, L"%_build" );
	ExpandVariables( temp, 1 );
	ShellVerBuild = wcstol( temp, &sep1, 10 );

	if ( ( ShellVerMajor < 7 ) || ( ShellVerMajor > 99 ) || ( ShellVerMinor < 0 ) || ( ShellVerMinor > 99 ) || ( ShellVerBuild < 0 ) || ( ShellVerBuild > 999 ) ) {
		ShellVerMajor = ShellVerMinor = ShellVerBuild = -1;
		return -1;
	}

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/fortune.html";
	piInfo.pszDescription	= (LPTSTR) L"Fortune cookie plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



int SeedRandomGenerator( )
{
	DWORD q = 0, a = 0, b = 0;
	SYSTEMTIME time;

	GetSystemTime( &time );
	a = ( ( time.wHour << 16 ) | time.wMinute ) * 1997;
	b = ( ( time.wSecond << 16 ) | time.wMilliseconds ) * 449;
	q = ( ( time.wMonth << 16 ) | time.wDay ) * 17;
	q = q ^ a ^ b;

	srand( q );
	RandomCount = 0;
	return 0;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	SeedRandomGenerator();
	GetShellVersion();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



int ParseArgs( LPTSTR inString, int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 7 - force uppercase
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & 3 ) == 0 )
		flags |= 1;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ch == '"' ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & 8 ) )
				ch = ':';
			else
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & 33 ) == 33 ) )
			flags &= 0xfffffe;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & 16 ) && ( arg[argnum][0] == '=' ) ) {
			continue;
		}

		if ( iswspace( ch ) && !quote && ( flags & 1 ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & 2 ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & 4 ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( flags & 16 ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= 0xfffffe;
			//- Printf( L"   * kludge invoked; eqkl = %d\n", eqkl );
		}
		else if ( ( ch == '=' ) && !quote && ( flags & 16 ) ) {
			eqkl = -1;
			flags &= 0xfffffe;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & 128 )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	for ( i = 0; i < argnum; i++ ) {
		while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
			arglen[i]--;
			arg[i][arglen[i]] = '\0';
		}
	}

	numargs = argnum;
	return argnum;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsTerminus( wchar_t ch )
{
	return ( ( ch == CR ) || ( ch == LF ) || ( ch == '\0' ) );
}



BOOL IsFraction( wchar_t ch )
{
	if ( ( ch >= 0xbc ) && ( ch <= 0xbe ) )
		return TRUE;
	else if ( ( ch >= 0x2150 ) && ( ch <= 0x215e ) )
		return TRUE;
	else
		return FALSE;
}



BOOL IsVowel2( wchar_t ch )
{
	ch = towlower( ch );
	if ( ( ch == 'a' ) || ( ch == 'e' ) || ( ch == 'i' ) || ( ch == 'o' ) || ( ch == 'u' ) )
		return TRUE;

	if ( ( ch == 0x0100 ) || ( ch == 0x0101 ) || ( ch == 0x0112 ) || ( ch == 0x0113 ) || ( ch == 0x012a ) || ( ch == 0x012b ) || ( ch == 0x014c ) || ( ch == 0x014d ) || ( ch == 0x016a ) || ( ch == 0x016b ) )
		return TRUE;		//	vowels with macrons -- used in hawai'ian

	return FALSE;
}



int ParseInt( LPTSTR inString, unsigned int *outValue )
{
	if ( ( inString == NULL ) || ( outValue == NULL ) )
		return -1;

	unsigned long int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}

	return -1;
}



int ParseInt( LPTSTR inString, unsigned long int *outValue )
{
	if ( ( inString == NULL ) || ( outValue == NULL ) )
		return -1;

	unsigned long int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}

	return -1;
}



int GetRand( int Range, BOOL Biased )
{
	unsigned int RandVal = 0;
	double real = 0.0;

	if ( !rand_s( &RandVal ) )
		real = (double) RandVal / ( UINT_MAX + 1.0 );
	else {
		real = (double) ( rand() | ( rand() << 5 ) ) / ( ( RAND_MAX + 1.0 ) * 32.0 );
		if ( RandomCount++ >= 3000 )
			SeedRandomGenerator();
	}

	if ( Biased )
		 real = pow( real, 1.15 );

	real = real * (double) Range;
	return (int) real;
}



unsigned int ExcludePriors( unsigned int inLine, DWORD PriorLine[], int TestPriors )
{
	if ( !TestPriors )
		return inLine;

	if ( ( inLine >= MAX_LINES ) || ( PriorLine == NULL ) || ( TestPriors < 0 ) || ( TestPriors > MAX_PRIORS ) )
		return UINT_MAX;

	if ( ( inLine + TestPriors ) >= MAX_LINES )
		return UINT_MAX;

	unsigned char Used[MAX_LINES];
	int i = 0, j = 0;

	memset( Used, 0x00, sizeof( Used ) );

	for ( i = 0; i < TestPriors; i++ )
		if ( PriorLine[i] < MAX_LINES )
			Used[PriorLine[i]] = 0xff;

	for ( i = 0; i < MAX_LINES; i++ ) {
		if ( Used[i] == 0 ) {
			if ( j == inLine )
				return i;

			j++;
		}
	}

	DisplayErrorHeader();
	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"   * Error:  Fell off the end of Used[]\r\n" );
	return UINT_MAX;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int StringSize( LPTSTR inString )
{
	if ( inString == NULL )
		return 0;

	return ( (int) wcslen( inString ) + 1 ) * sizeof( wchar_t );
}



BOOL FoundInMulti( wchar_t *MultiString, LPTSTR inString )
{
	if ( ( MultiString == NULL ) || ( inString == NULL ) )
		return FALSE;

	if ( ( MultiString[0] == '\0' ) || ( inString == '\0' ) )
		return FALSE;

	int i = 0;

	while ( *( MultiString + i ) != '\0' ) {
		if ( !_wcsicmp( MultiString + i, inString ) )
			return TRUE;

		i += (int) wcslen( MultiString + i ) + 1;
	}

	return FALSE;
}



int AppendChar( LPTSTR outString, int MaxLength, wchar_t ch )
{
	if ( ( outString == NULL ) || ( MaxLength < 2 ) )
		return -1;

	int i = 0;

	while ( outString[i] != '\0' ) {
		i++;
		if ( i >= ( MaxLength - 1 ) )
			return -1;
	}

	outString[i++] = ch;
	outString[i] = '\0';
	return 0;
}



BOOL IsAFile( LPTSTR Filename )
{
	if ( Filename == NULL )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;

	WIN32_FILE_ATTRIBUTE_DATA FileInfo;

	if ( !GetFileAttributesEx( Filename, GetFileExInfoStandard, &FileInfo ) )
		return FALSE;

	if ( ( FileInfo.dwFileAttributes & FILE_ATTRIBUTE_DEVICE ) || ( FileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) || ( FileInfo.dwFileAttributes & FILE_ATTRIBUTE_VIRTUAL ) )
		return FALSE;

	if ( ( FileInfo.ftCreationTime.dwHighDateTime == 0 ) && ( FileInfo.ftCreationTime.dwLowDateTime == 0 ) && ( FileInfo.ftLastWriteTime.dwHighDateTime == 0 ) && ( FileInfo.ftLastWriteTime.dwLowDateTime == 0 ) && ( FileInfo.nFileSizeLow == 0 ) && ( FileInfo.nFileSizeHigh == 0 ) )
		return FALSE;

	return TRUE;
}



int StripQuotes( LPTSTR inString )
{
	if ( inString == NULL )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;

	wchar_t TempBuffer[MAX_ARG_LEN] = L"";
	int i = 0, j = 0, Count = 0;
	wchar_t ch = '\0';

	do {
		ch = inString[i++];

		if ( ch == '"' )
			Count++;
		else if ( j < ( MAX_ARG_LEN - 1 ) ) {
			TempBuffer[j++] = ch;
			TempBuffer[j] = '\0';
		}

	} while ( ch != '\0' );

	if ( Count )
		wcscpy_s( inString, MAX_ARG_LEN, TempBuffer );

	return Count;
}



BOOL IsWordBeginningWithApostrophe( LPTSTR inString )
{
	if ( !inString )
		return FALSE;
	else if ( inString[0] == '\0' )
		return FALSE;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	wchar_t ch = '\0';
	int i = 0;

	const int NumWords = 21;
	const wchar_t Word[NumWords][12] = { L"'cos", L"'cus", L"'em", L"'mid", L"'mongst",
		L"'n", L"'n'", L"'neath", L"'tain't", L"'til", L"'tis", L"'tisn't", L"'twas", L"'twasn't",
		L"'tween", L"'twere", L"'tweren't", L"'twixt", L"'twould", L"'twouldn't", L"'un" };

	do {
		ch = towlower( inString[i] );
		if ( ( ch == 0x2018 ) || ( ch == 0x2019 ) )
			ch = APOS;

		if ( !iswalpha( ch ) && ( ch != APOS ) )
			ch = '\0';

		if ( i < ( SHORT_BUF_LEN - 1 ) )
			Temp[i++] = ch;

		Temp[i] = '\0';

	} while ( ch && ( i < ( SHORT_BUF_LEN - 1 ) ) );

	for ( i = 0; i < NumWords; i++ )
		if ( !wcscmp( Temp, Word[i] ) )
			return TRUE;

	return FALSE;
}



int GetEnvironmentChar( LPTSTR Name, wchar_t *outChar )
{
	if ( !Name )
		return -1;
	else if ( Name[0] == '\0' )
		return 0;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	unsigned int NumericChar = 0;
	int i = 0;

	if ( GetEnvironmentVariable( Name, Temp, SHORT_BUF_LEN ) < 1 )
		return 0;
	else if ( Temp[0] == '\0' )
		return 0;

	if ( ( Temp[0] > 31 ) && ( Temp[1] == '\0' ) )
		if ( !iswdigit( Temp[0] ) ) {
			*outChar = Temp[0];
			return 0;
		}

	while ( iswspace( Temp[i] ) )
		i++;

	if ( !iswdigit( Temp[i] ) )
		return -1;

	if ( ParseInt( Temp + i, &NumericChar ) )
		return -1;

	if ( ( NumericChar < 32 ) || ( NumericChar > 0xffff ) )
		return -1;

	*outChar = NumericChar;
	return 0;
}



BOOL IsEnd( wchar_t ch )
{
	if ( ( ch == '\0' ) || ( ch == CR ) || ( ch == LF ) || ( ch == PILCROW ) || ( ch == SECTION_MARK ) )
		return TRUE;
	else
		return FALSE;
}



int FixQuotes( LPTSTR inString )
{
	if ( inString == NULL )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;

	wchar_t ch = 0x20, lastch = '\0', nextch = '\0';
	int i = 0, j = 0;
	wchar_t Temp[MAX_ARG_LEN] = L"";
	BOOL QuoteToggle = FALSE, Changes = FALSE;

	wchar_t OpenQuote = LQUOTE,		CloseQuote = RQUOTE;
	wchar_t OpenSQuote = LSQUOTE,	CloseSQuote = RSQUOTE;
	wchar_t Apostrophe = RSQUOTE,	Okina = LSQUOTE;
	wchar_t Prime = APOS,			DoublePrime = QUOTE;

	GetEnvironmentChar( L"OPENQUOTE",   &OpenQuote );
	GetEnvironmentChar( L"CLOSEQUOTE",  &CloseQuote );
	GetEnvironmentChar( L"OPENSQUOTE",  &OpenSQuote );
	GetEnvironmentChar( L"CLOSESQUOTE", &CloseSQuote );
	GetEnvironmentChar( L"APOSTROPHE",  &Apostrophe );
	GetEnvironmentChar( L"PRIME",       &Prime );
	GetEnvironmentChar( L"DOUBLEPRIME", &DoublePrime );
	GetEnvironmentChar( L"'OKINA",		&Okina );


	while ( inString[i] ) {
		lastch = ch;
		ch = inString[i];
		if ( ch )
			nextch = inString[i + 1];
		else
			nextch = '\0';

		if ( ( ch == QUOTE ) && ( lastch != '\\' ) ) {
			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || IsTerminus( nextch ) || ( nextch == PILCROW ) || ( nextch == SECTION_MARK ) ) )
				ch = DoublePrime;
			else if ( QuoteToggle ) {
				ch = CloseQuote;
				QuoteToggle = FALSE;
			}
			else if ( IsEnd( nextch ) && !IsEnd( lastch ) ) {
				ch = CloseQuote;
				QuoteToggle = FALSE;
			}
			else {
				ch = OpenQuote;
				QuoteToggle = TRUE;
			}
		}
		else if ( ( ch == APOS ) && ( lastch != '\\' ) ) {
			if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
				ch = CloseSQuote;

				if ( iswalnum( lastch ) && iswalpha( nextch ) )
					ch = Apostrophe;
			}
			else {
				ch = OpenSQuote;

				if ( iswspace( lastch ) )
					if ( IsTerminus( nextch ) || ( nextch == PILCROW ) || ( nextch == SECTION_MARK ) )
						ch = CloseSQuote;
			}

			if ( !iswalpha( lastch ) && iswalpha( nextch ) ) {
				if ( IsWordBeginningWithApostrophe( inString + i ) )
					ch = Apostrophe;
			}
			else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
				ch = Okina;

			if ( !iswalnum( lastch ) && iswdigit( nextch ) )
				if ( iswdigit( inString[i + 2] ) )
					if ( !iswalnum( inString[i + 3] ) )
						ch = Apostrophe;

			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || IsTerminus( nextch ) || ( nextch == PILCROW ) || ( nextch == SECTION_MARK ) ) )
				ch = Prime;
		}
		else if ( ( ch == '-' ) && ( nextch == '-' ) && ( lastch != '-' ) ) {
			if ( inString[i + 2] != '-' ) {
				ch = EMDASH;
				i++;
			}
		}
		else if ( IsTerminus( ch ) || ( ch == PILCROW ) || ( ch == SECTION_MARK ) )
			QuoteToggle = FALSE;

		Changes |= ( ch != inString[i] );
		i++;
		Temp[j++] = ch;
		Temp[j] = '\0';
	}

	if ( Changes )
		wcscpy_s( inString, MAX_ARG_LEN, Temp );

	return 0;
}



int OutputWidth( )
{
	if ( !QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) ) )
		return 0;

	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Width = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			Width = ConsoleInfo.dwSize.X;
	}

	return Width;
}



int OutputIsUnicode( )
{
	if ( QueryUnicodeOutput() )
		return 16;

	if ( ShellVerMajor < 13 )
		return 0;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor < 4 ) )
		return 0;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor == 4 ) && ( ShellVerBuild < 46 ) )
		return 0;

	wchar_t temp[SHORT_BUF_LEN] = L"%@option[utf8output]";
	ExpandVariables( temp, 1 );

	if ( !_wcsicmp( temp, L"yes" ) )
		return 8;
	else
		return 0;
}



int ExpandTabs( LPTSTR inString, int n, BOOL WrapAtPilcrows )
{
	if ( ( inString == NULL ) || ( n < 1 ) || ( n > MAX_TAB_SIZE ) )
		return -1;

	if ( inString[0] == '\0' )
		return -1;

	wchar_t TempBuffer[MAX_ARG_LEN] = L"";
	wchar_t ch = 0xff;
	int i = 0, j = 0, k = 0, l = 0, q = 0;

	for ( i = 0; ch != '\0'; i++ ) {
		ch = inString[i];

		if ( ch == 0x09 ) {
			k = n - ( q % n );
			for ( l = 0; l < k; l++ )
				if ( j < ( MAX_ARG_LEN - 1 ) )
					TempBuffer[j++] = 0x20;

			q += k;
		}
		else if ( j < ( MAX_ARG_LEN - 1 ) ) {
			TempBuffer[j++] = ch;
			q++;
		}
		TempBuffer[j] = '\0';

		if ( IsTerminus( ch ) || ( WrapAtPilcrows && ( ch == PILCROW ) ) )
			q = 0;
		else if ( ch == SECTION_MARK )
			q = 3;
	}

	if ( k )
		wcscpy_s( inString, MAX_ARG_LEN, TempBuffer );

	return 0;
}



int RemoveAttributions( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	wchar_t *StartAttrib = NULL, *EndAttrib = NULL;
	wchar_t ch = '\0';

	do {
		StartAttrib = wcschr( String, SECTION_MARK );
		if ( StartAttrib == NULL )
			return 0;

		EndAttrib = wcschr( StartAttrib, PILCROW );
		if ( !EndAttrib ) {
			*StartAttrib = '\0';
			return 0;
		}

		EndAttrib++;

		do {
			ch = *EndAttrib;
			EndAttrib++;

			*StartAttrib = ch;
			StartAttrib++;

		} while ( ch );

	} while ( TRUE );

	return 0;
}



int ExpandAttributions( LPTSTR String, BOOL EmDash )
{
	if ( String == NULL )
		return -1;

	wchar_t temp[MAX_ARG_LEN] = L"";

	int i = 0;
	wchar_t ch = '\0';

	while ( String[i] ) {
		ch = String[i++];

		if ( ch == SECTION_MARK ) {
			if ( EmDash )
				wcscat_s( temp, MAX_ARG_LEN, L"\x00b6\x00b6                    \x2014 " );
			else
				wcscat_s( temp, MAX_ARG_LEN, L"\x00b6\x00b6                    -- " );
		}
		else
			AppendChar( temp, MAX_ARG_LEN, ch );
	};

	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}



int LineWrap( LPTSTR inString, int Width )
{
	if ( inString == NULL )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;


	if ( !Width )
		return 0;


	wchar_t TempBuffer[MAX_ARG_LEN] = L"";
	wchar_t ch = '\0';
	int inBase = 0, index = 0, LastSpace = -1, LastHyphen = -1;

	do {
		index = 0;
		LastSpace = -1;
		LastHyphen = -1;

		do {
			ch = inString[inBase + index];

			if ( ( ch == '\0' ) || ( ch == PILCROW ) )
				break;

			if ( iswspace( ch ) && ( ch != NBSP ) ) {
				if ( inString[inBase + index + 1] != '-' ) {
					LastSpace = index;
					LastHyphen = -1;
				}
				else if ( ( inString[inBase + index + 2] != '-' ) && !iswspace( inString[inBase + index + 2] ) ){
					LastSpace = index;
					LastHyphen = -1;
				}
			}

			if ( ( index > 0 ) && ( index < ( Width - 1 ) ) && ( ch == '-' ) )
				if ( iswalpha( inString[inBase + index - 1] ) && iswalpha( inString[inBase + index + 1] ) )
					LastHyphen = index;

			index++;
		} while ( index < Width );

		if ( ch == '\0' ) {
			//  copy remainder of string

			wcscat_s( TempBuffer, MAX_ARG_LEN, inString + inBase );
		}
		else if ( ch == PILCROW ) {
			//  insert line break

			for ( index = 0; ( inString[inBase + index] != PILCROW ) && ( inString[inBase + index] != SECTION_MARK ); index++ )
				AppendChar( TempBuffer, MAX_ARG_LEN, inString[inBase + index] );

			AppendChar( TempBuffer, MAX_ARG_LEN, PILCROW );
			inBase  += index + 1;
		}
		else if ( LastSpace < ( Width / 2 ) ) {
			//  copy Width chars

			for ( index = 0; index < Width; index++ )
				AppendChar( TempBuffer, MAX_ARG_LEN, inString[inBase + index] );

			inBase  += Width;

			if ( inString[inBase] == PILCROW )
				inBase++;

			while ( iswspace( inString[inBase] ) )
				inBase++;
		}
		else {
			//  insert line break

			if ( ( LastHyphen >= ( Width / 2 ) ) && ( LastHyphen > ( LastSpace + 2 ) ) )
				LastSpace = LastHyphen + 1;

			for ( index = 0; index < LastSpace; index++ )
				AppendChar( TempBuffer, MAX_ARG_LEN, inString[inBase + index] );

			inBase  += LastSpace;
			AppendChar( TempBuffer, MAX_ARG_LEN, PILCROW );

			while ( iswspace( inString[inBase] ) )
				inBase++;
		}

	} while ( ch != '\0' );

	wcscpy_s( inString, MAX_ARG_LEN, TempBuffer );
	return 0;
}



int SubstituteNewlines( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	wchar_t temp[MAX_ARG_LEN] = L"";

	int i = 0, j = 0;
	wchar_t ch = '\0';

	while ( String[i] ) {
		ch = String[i++];

		if ( ( ch == PILCROW ) || ( ch == SECTION_MARK ) ) {
			if ( j < ( MAX_ARG_LEN - 1 ) )
				temp[j++] = CR;

			if ( j < ( MAX_ARG_LEN - 1 ) )
				temp[j++] = LF;

			temp[j] = '\0';
		}
		else {
			if ( j < ( MAX_ARG_LEN - 1 ) )
				temp[j++] = ch;

			temp[j] = '\0';
		}
	};

	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}



int UpdateMRUList( LPTSTR Filename, BOOL Delete, HKEY PluginKeyHandle )
{
	if ( Filename == NULL )
		return -1;

	wchar_t *OldMRU = NULL, *NewMRU = NULL;
	DWORD OldMRUSize = 0, NewMRUSize = 0;
	DWORD OldMRUType = 0, NewMRUType = 0;
	int i = 0, j = 0;
	int k = 0, len = 0, rv = 0;

	wchar_t ValueName[MAX_REG_VALUE_NAME_LEN] = L"";
	DWORD ValueNameLen = MAX_REG_VALUE_NAME_LEN;
	DWORD ValueType = 0;


	rv = RegQueryValueEx( PluginKeyHandle, MRUValueName, NULL, &OldMRUType, NULL, &OldMRUSize );

	if ( ( rv == ERROR_SUCCESS ) && ( OldMRUType == REG_MULTI_SZ ) ) {
		OldMRU = (wchar_t *) malloc( OldMRUSize );
		RegQueryValueEx( PluginKeyHandle, MRUValueName, NULL, NULL, (BYTE *) OldMRU, &OldMRUSize );
	}

	NewMRUSize = StringSize( Filename ) + OldMRUSize + 16;
	NewMRU = (wchar_t *) malloc( NewMRUSize );

	if ( !Delete ) {
		wcscpy_s( NewMRU, ( NewMRUSize / sizeof( wchar_t ) ), Filename );
		i = (int) wcslen( Filename ) + 1;
		k = 1;
	}

	if ( ( OldMRU != NULL ) && ( OldMRUSize > 0 ) ) {
		while ( OldMRU[j] != '\0' ) {
			len = (int) wcslen( OldMRU + j ) + 1;
			if ( ( _wcsicmp( OldMRU + j, Filename ) != 0 ) && ( k < MAX_MRU_ENTRIES ) ) {
				wcscpy_s( NewMRU + i, ( NewMRUSize / sizeof( wchar_t ) ) - i, OldMRU + j );
				i += len;
				k++;
			}
			j += len;
		};
	}

	*( NewMRU + i++ ) = '\0';

	if ( k )
		RegSetValueEx( PluginKeyHandle, MRUValueName, 0, REG_MULTI_SZ, (BYTE *) NewMRU, i * sizeof( wchar_t ) );
	else
		RegDeleteValue( PluginKeyHandle, MRUValueName );

	
	i = 0;

	while ( RegEnumValue( PluginKeyHandle, i, ValueName, &ValueNameLen, NULL, &ValueType, NULL, NULL ) == ERROR_SUCCESS ) {
		ValueNameLen = MAX_REG_VALUE_NAME_LEN;
		i++;

		if ( !_wcsicmp( ValueName, MRUValueName ) && ( ValueType == REG_MULTI_SZ ) )
			continue;

		if ( ValueType == REG_BINARY )
			if ( k )
				if ( FoundInMulti( NewMRU, ValueName ) )
					continue;

		RegDeleteValue( PluginKeyHandle, ValueName );
		i = 0;
	}


	if ( OldMRU != NULL )
		free( OldMRU );

	free( NewMRU );
	return 0;
}



int ByteFromFile( HANDLE FileHandle, LPTSTR Filename, BYTE *Buffer, DWORD BufSize, DWORD *BytesRead, DWORD *BufIndex )
{
	if ( ( FileHandle == NULL ) || ( Filename == NULL ) || ( Buffer == NULL ) || ( BytesRead == NULL ) || ( BufIndex == NULL ) )
		return -1;

	if ( BufSize < 512 )
		return -1;

	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	int err = 0;
	int rv = 0;


	if ( *BufIndex >= *BytesRead ) {

		if ( !ReadFile( FileHandle, Buffer, BufSize, BytesRead, NULL ) ) {
			err = GetLastError();
			GetErrorText( err, ErrorText );
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error reading \"%s\":\r\n", Filename );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s\r\n", ErrorText );
			free( Buffer );
			CloseHandle( FileHandle );
			*BytesRead = 0;
			*BufIndex = -1;
			return -1;
		}
		else if ( *BytesRead == 0 ) {
			*BufIndex = -2;
			return -2;
		}

		*BufIndex = 0;
	}

	rv = Buffer[*BufIndex];
	*BufIndex += 1;
	return rv;
}



wchar_t CharFromFile( HANDLE FileHandle, LPTSTR Filename, int FileEncoding, unsigned int CodePageIndex, BYTE *Buffer, DWORD BufSize, DWORD *BytesRead, DWORD *BufIndex )
{
	if ( ( FileHandle == NULL ) || ( Filename == NULL ) || ( Buffer == NULL ) || ( BytesRead == NULL ) || ( BufIndex == NULL ) )
		return -1;

	if ( BufSize < 512 )
		return -1;

	int b1 = ByteFromFile( FileHandle, Filename, Buffer, BufSize, BytesRead, BufIndex );
	if ( b1 < 0 )
		return 0;

	if ( FileEncoding == FILE_UTF16 ) {
		int b2 = ByteFromFile( FileHandle, Filename, Buffer, BufSize, BytesRead, BufIndex );
		if ( b2 < 0 )
			return 0;

		return ( b1 | ( b2 << 8 ) );
	}
	else if ( FileEncoding == FILE_UTF8 ) {
		if ( b1 < 0x80 )
			return b1;
		else if ( b1 < 0xc0 )
			return INVALID_CHAR;
		else if ( b1 < 0xe0 ) {
			int b2 = ByteFromFile( FileHandle, Filename, Buffer, BufSize, BytesRead, BufIndex );
			if ( b2 < 0 )
				return 0;

			if ( ( b2 & 0xc0 ) != 0x80 )
				return INVALID_CHAR;
			else
				return ( ( b1 & 0x1f ) << 6 ) | ( b2 & 0x3f );
		}
		else if ( b1 < 0xf0 ) {
			int b2 = ByteFromFile( FileHandle, Filename, Buffer, BufSize, BytesRead, BufIndex );;
			if ( b2 < 0 )
				return 0;

			int b3 = ByteFromFile( FileHandle, Filename, Buffer, BufSize, BytesRead, BufIndex );
			if ( b3 < 0 )
				return 0;

			if ( ( ( b2 & 0xc0 ) != 0x80 ) || ( ( b3 & 0xc0 ) != 0x80 ) )
				return INVALID_CHAR;
			else
				return ( ( b1 & 0x0f ) << 12 ) | ( ( b2 & 0x3f ) << 6 ) | ( b3 & 0x3f );
		}
		else
			return INVALID_CHAR;
	}
	else {
		if ( b1 & 0x80 )
			return CodePageTable[CodePageIndex][b1 & 0x7f];
		else
			return b1;
	}
}



unsigned int LooksLikeUTF8( BYTE *Buffer, unsigned int Bytes )
{
	if ( Buffer == NULL )
		return 0;

	if ( Bytes < 16 )
		return 0;

	BYTE b = 0;
	int c = 0, mbs = 0;


	for ( unsigned int i = 0; i < Bytes; i++ ) {
		b = Buffer[i];

		if ( c ) {
			if ( ( b & 0xc0 ) != 0x80 )
				return 0;

			c--;
		}
		else {
			if ( b < 0x80 )
				;
			else if ( b < 0xc0 )
				return 0;
			else if ( b < 0xe0 ) {
				c = 1;
				mbs++;
			}
			else if ( b < 0xf0 ) {
				c = 2;
				mbs++;
			}
			else if ( b < 0xf8 ) {
				c = 3;
				mbs++;
			}
			else
				return 0;
		}
	}

	return mbs;
}



int RandomLine( LPTSTR inFilename, LPTSTR outString, int Priors, BOOL Biased, BOOL UpdateRegistry, unsigned long int TestLine, unsigned int CodePageIndex )
{
	if ( !inFilename || !outString )
		return -1;

	wchar_t Filename[MAX_FILENAME_LEN] = L"";

	HANDLE FileHandle = NULL;
	BYTE *Buffer = NULL;
	DWORD BytesRead = 0, FilePointer = 0, index = 0;
	int FileEncoding = FILE_OEM;

	unsigned long int NumLines = 0;
	unsigned long int LineOffset[MAX_LINES];

	HKEY PluginKeyHandle = NULL;
	DWORD DataType = 0, DataSize = 0;
	DWORD PriorLine[MAX_PRIORS];
	DWORD TestPriors = 0;

	int Search = 0;
	unsigned long int LineStart = ULONG_MAX;
	wchar_t ch = '\0';

	int err = 0;
	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	int rv = 0;
	unsigned long int LineNumber = 0;
	DWORD savindex = 0;


	QueryTrueName( inFilename, Filename );


	if ( ( Priors < 0 ) || ( Priors > MAX_PRIORS ) )
		Priors = DEFAULT_PRIORS;


	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( FileHandle == INVALID_HANDLE_VALUE ) {
		err = GetLastError();
		GetErrorText( err, ErrorText );
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error opening \"%s\":\r\n", Filename );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s\r\n", ErrorText );
		return 2;
	}

	if ( GetFileType( FileHandle ) != FILE_TYPE_DISK ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Not a file: \"%s\"\r\n", Filename );
		CloseHandle( FileHandle );
		return 2;
	}


	Buffer = (BYTE*) malloc( BUFFER_SIZE );

	if ( Buffer == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error allocating file buffer\r\n" );
		CloseHandle( FileHandle );
		return 2;
	}

	ch = CharFromFile( FileHandle, Filename, 0, 0, Buffer, BUFFER_SIZE, &BytesRead, &index );

	if ( index == -1 )
		return 2;
	else if ( index == -2 ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"\"%s\" is empty!\r\n", Filename );
		free( Buffer );
		CloseHandle( FileHandle );
		return 2;
	}

	if ( BytesRead >= 2 )
		if ( IsTextUnicode( Buffer, BytesRead, NULL ) )
			FileEncoding = FILE_UTF16;

	if ( FileEncoding == FILE_UTF16 ) {
		if ( ( Buffer[0] == 0xff ) && ( Buffer[1] == 0xfe ) )
			FilePointer = 2;
	}
	else if ( ( Buffer[0] == 0xef ) && ( Buffer[1] == 0xbb ) && ( Buffer[2] == 0xbf ) ) {
		FilePointer = 3;
		FileEncoding = FILE_UTF8;
	}
	else if ( LooksLikeUTF8( Buffer, BytesRead ) )
		FileEncoding = FILE_UTF8;

	index = FilePointer;

	do {
		savindex = index;
		ch = CharFromFile( FileHandle, Filename, FileEncoding, CodePageIndex, Buffer, BUFFER_SIZE, &BytesRead, &index );

		if ( index == -1 )
			return 2;
		else if ( index == -2 )
			Search = 2;

		if ( Search == 0 ) {
			if ( !IsTerminus( ch ) ) {
				if ( LineStart == ULONG_MAX )
					LineStart = FilePointer;

				if ( !iswspace( ch ) ) {
					if ( NumLines < MAX_LINES ) {
						LineOffset[NumLines++] = LineStart;
						Search = 1;
					}
					else
						Search = 2;
				}
			}
			else
				LineStart = ULONG_MAX;
		}
		else if ( Search == 1 ) {
			if ( IsTerminus( ch ) ) {
				Search = 0;
				LineStart = ULONG_MAX;
			}
		}

		if ( index > savindex )
			FilePointer += ( index - savindex );
		else
			FilePointer += ( index + BUFFER_SIZE - savindex );

	} while ( Search < 2 );

	if ( NumLines == 0 ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"\"%s\" contains no text!\r\n", Filename );
		free( Buffer );
		CloseHandle( FileHandle );
		return 2;
	}


	//  get list of priors

	RegCreateKeyEx( HKEY_CURRENT_USER, PluginKeyName, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &PluginKeyHandle, NULL );

	memset( PriorLine, 0xff, sizeof( PriorLine ) );

	if ( PluginKeyHandle != NULL ) {
		DataSize = sizeof( PriorLine );
		RegQueryValueEx( PluginKeyHandle, Filename, NULL, &DataType, (BYTE *) &PriorLine, &DataSize );
		if ( DataType != REG_BINARY )
			memset( PriorLine, 0xff, sizeof( PriorLine ) );
	}

	TestPriors = Priors;

	if ( NumLines < 4 )
		TestPriors = 0;
	else if ( TestPriors > ( NumLines - 4 ) )
		TestPriors = NumLines - 4;


	if ( TestLine >= NumLines ) {

		LineNumber = GetRand( NumLines - TestPriors, Biased );

		if ( TestPriors )
			LineNumber = ExcludePriors( LineNumber, PriorLine, TestPriors );
	}
	else {
		LineNumber = TestLine;
		UpdateRegistry = FALSE;
	}


	SetFilePointer( FileHandle, LineOffset[LineNumber], NULL, 0 );
	index = BytesRead = 0;
	outString[0] = '\0';
	int j = 0;

	do {
		ch = CharFromFile( FileHandle, Filename, FileEncoding, CodePageIndex, Buffer, BUFFER_SIZE, &BytesRead, &index );

		if ( index == -1 )
			return 2;
		else if ( IsTerminus( ch ) )
			break;

		outString[j++] = ch;
		outString[j] = '\0';

	} while ( j < ( MAX_ARG_LEN - 1 ) );


	//  update list of priors

	for ( int i = ( MAX_PRIORS - 1 ); i > 0; i-- )
		PriorLine[i] = PriorLine[i - 1];

	PriorLine[0] = LineNumber;

	Search = 0;
	while ( ( Search < Priors ) && ( PriorLine[Search] != UINT_MAX ) )
		Search++;

	if ( UpdateRegistry && ( PluginKeyHandle != NULL ) ) {
		if ( Priors )
			RegSetValueEx( PluginKeyHandle, Filename, 0, REG_BINARY, (BYTE *) &PriorLine, Search * sizeof( DWORD ) );
		else
			RegDeleteValue( PluginKeyHandle, Filename );

		UpdateMRUList( Filename, ( Priors == 0 ), PluginKeyHandle );
	}

	if ( UpdateRegistry )
		LastLine = LineNumber;


	//  clean up

	free( Buffer );
	CloseHandle( FileHandle );
	if ( PluginKeyHandle != NULL )
		RegCloseKey( PluginKeyHandle );

	return 0;
}



DLLExports int WINAPI fortune( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN] = L"";
	unsigned int Priors = UINT_MAX;
	BOOL Biased				= FALSE;
	BOOL HideAttributions	= FALSE;
	BOOL DoQuotes			= FALSE;
	BOOL NoBOM				= FALSE;
	BOOL NoSpecialChars		= FALSE;
	BOOL LineFeed			= FALSE;
	unsigned int TabSize = 4;
	wchar_t outString[MAX_ARG_LEN] = L"";
	unsigned int Width = OutputWidth();
	unsigned long int TestLine = ULONG_MAX;
	unsigned int CodePage = 0, CodePageIndex = 0;
	int rv = 0;


	ParseArgs( inString, 133 );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) ) {
				QPuts( (LPTSTR) FortunePurposeText );
				QPuts( (LPTSTR) FortuneUsageText );
				return 0;
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				QPuts( (LPTSTR) FortunePurposeText );
				QPuts( (LPTSTR) FortuneUsageText );
				Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );
				return 0;
			}
			else if ( !wcscmp( arg[i], L"/B" ) )
				Biased = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &CodePage ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad code page \"%s\"\r\n", arg[i] );
					return 1;
				}
				
				CodePageIndex = -1;
				for ( j = 0; j < NUM_CODE_PAGES; j++ )
					if ( CodePageNumber[j] == CodePage )
						CodePageIndex = j;

				if ( CodePageIndex == -1 ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Code page %u not supported\r\n", CodePage );
					return 1;
				}
			}
			else if ( !wcscmp( arg[i], L"/H" ) )
				HideAttributions = TRUE;
			else if ( !wcscmp( arg[i], L"/LF" ) )
				LineFeed = TRUE;
			else if ( !_wcsicmp( arg[i], L"/NB" ) )
				NoBOM = TRUE;
			else if ( !_wcsicmp( arg[i], L"/NC" ) )
				NoSpecialChars = TRUE;
			else if ( !wcsncmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Priors ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( Priors > MAX_PRIORS ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"n must be 0 - %u\r\n", MAX_PRIORS );
					return 1;
				}
			}
			else if ( !wcscmp( arg[i], L"/Q" ) )
				DoQuotes = TRUE;
			else if ( !wcsncmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &TabSize ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( TabSize > MAX_TAB_SIZE ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Tab size must be 0 - %u\r\n", MAX_TAB_SIZE );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Width ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( ( Width > MAX_COLUMNS ) || ( ( Width != 0 ) && ( Width < MIN_COLUMNS ) ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Width must be 0 or %u - %u\r\n", MIN_COLUMNS, MAX_COLUMNS );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/#", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &TestLine ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option: \"%s\"\r\n", arg[i] );
					return 1;
			}
			else if ( Filename[0] == '\0' )
				wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] );
			else {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( CodePage == 0 ) {
		CodePageIndex = 0;
		CodePage = GetACP();

		for ( int k = 0; k < NUM_CODE_PAGES; k++ )
			if ( CodePageNumber[k] == CodePage )
				CodePageIndex = k;
	}

	if ( Filename[0] == '\0' )
		GetEnvironmentVariable( EnvVarName, Filename, MAX_FILENAME_LEN );

	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing filename\r\n" );
		return 1;
	}


	if ( ( Filename[0] == '@' ) && Filename[1] ) {
		if ( IsAFile( Filename + 1 ) ) {
			
			rv = RandomLine( Filename + 1, Filename, 0, FALSE, FALSE, ULONG_MAX, CodePageIndex );
			if ( rv || ( Filename[0] == '\0' ) )
				return 2;

			StripQuotes( Filename );
		}
	}

	if ( TestLine != ULONG_MAX )
		Priors = 0;

	if ( Priors == UINT_MAX )
		Priors = DEFAULT_PRIORS;


	rv = RandomLine( Filename, outString, Priors, Biased, TRUE, TestLine, CodePageIndex );

	if ( rv == 0 ) {
		wcscpy_s( LastFortune, MAX_ARG_LEN, outString );
		ExpandTabs( outString, TabSize, ( Width > 0 ) );

		if ( DoQuotes )
			FixQuotes( outString );

		if ( !NoSpecialChars ) {
			if ( HideAttributions )
				RemoveAttributions( outString );
			else
				ExpandAttributions( outString, DoQuotes );

			if ( Width )
				LineWrap( outString, Width );

			SubstituteNewlines( outString );
		}

		if ( !NoBOM )
			if ( QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) ) == 0 )
				if ( OutputIsUnicode() )
					QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );

		if ( LineFeed )
			QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), LF );

		Printf( L"%s\r\n", outString );
	}

	return rv;
}



DLLExports int WINAPI f_fortune( LPTSTR inString )
{
	if ( !inString )
		return -1;


	unsigned int Priors = UINT_MAX, flags = 0, Width = 0, TabSize = 0;
	BOOL Biased = FALSE;
	unsigned int CodePage = 0, CodePageIndex = 0;
	int rv = 0;


	ParseArgs( inString, 130 );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing filename\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Priors ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid n: \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( Priors > MAX_PRIORS ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"n must be 0 - %u\r\n", MAX_PRIORS );
			return -1;
		}
	}

	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid flags: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Width ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid width: \"%s\"\r\n", arg[3] );
			return -1;
		}

		if ( ( Width > MAX_COLUMNS ) || ( ( Width != 0 ) && ( Width < MIN_COLUMNS ) ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Width must be 0 or %u - %u\r\n", MIN_COLUMNS, MAX_COLUMNS );
			return -1;
		}

		if ( Width )
			TabSize = 4;
	}

	if ( arglen[4] ) {
		if ( ParseInt( arg[4], &TabSize ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid tab size: \"%s\"\r\n", arg[4] );
			return -1;
		}

		if ( TabSize > MAX_TAB_SIZE ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Tab size must be 0 - %u\r\n", MAX_TAB_SIZE );
			return -1;
		}
	}

	if ( arglen[5] ) {
		if ( ParseInt( arg[5], &CodePage ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad code page \"%s\"\r\n", arg[5] );
			return -1;
		}

		CodePageIndex = -1;
		for ( int k = 0; k < NUM_CODE_PAGES; k++ )
			if ( CodePageNumber[k] == CodePage )
				CodePageIndex = k;

		if ( CodePageIndex == -1 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Code page %u not supported\r\n", CodePage );
			return -1;
		}
	}


	if ( Priors == UINT_MAX )
		Priors = DEFAULT_PRIORS;

	if ( CodePage == 0 ) {
		CodePageIndex = 0;
		CodePage = GetACP();

		for ( int k = 0; k < NUM_CODE_PAGES; k++ )
			if ( CodePageNumber[k] == CodePage )
				CodePageIndex = k;
	}

	rv = RandomLine( arg[0], inString, Priors, ( flags & FLAG_BIASED ), TRUE, ULONG_MAX, 0 );
	
	if ( rv == 0 ) {
		ExpandTabs( inString, TabSize, ( Width > 0 ) );

		if ( flags & FLAG_FIX_QUOTES )
			FixQuotes( inString );

		if ( flags & FLAG_NO_ATTRIBS )
			RemoveAttributions( inString );

		if ( Width ) {
			ExpandAttributions( inString, ( flags & FLAG_FIX_QUOTES ) > 0 );
			LineWrap( inString, Width );
		}

		if ( flags & FLAG_RETURN_EOLS )
			SubstituteNewlines( inString );
	}

	if ( rv > 0 )
		return -rv;
	else
		return rv;
}



/*
DLLExports int WINAPI f_randb( LPTSTR inString )
{
	if ( !inString )
		return -1;

	unsigned long int Range = 0, flags = 0;
	int Random = 0;

	ParseArgs( inString, 130 );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Range ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid range: \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( ( Range < 2 ) || ( Range > 1048576 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid range: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing range\r\n", arg[0] );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &flags ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid flags: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	Random = GetRand( Range, ( flags & 1 ) );

	wsprintf( inString, L"%d", Random );
	return 0;
}
*/



DLLExports int WINAPI _lastfline( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wsprintf( outString, L"%d", LastLine );
	return 0;
}



DLLExports int WINAPI _lastfortune( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, MAX_ARG_LEN, LastFortune );
	return 0;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t helpbuf[MAX_ARG_LEN] = L"";

	if ( _wcsicmp( inString, L"fortune" ) == 0 ) {
		wcscpy_s( helpbuf, MAX_ARG_LEN, FortunePurposeText );
		wcscat_s( helpbuf, MAX_ARG_LEN, FortuneUsageText );
		MessageBox( GetForegroundWindow(), helpbuf, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( _wcsicmp( inString, L"@fortune" ) == 0 ) {
		MessageBox( GetForegroundWindow(), f_FortuneUsageText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( _wcsicmp( inString, L"_lastfline" ) == 0 ) {
		MessageBox( GetForegroundWindow(), _LastFLineUsageText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( _wcsicmp( inString, L"_lastfortune" ) == 0 ) {
		MessageBox( GetForegroundWindow(), _LastFortuneUsageText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"fortune" ) ) {
		wcscpy_s( outString, 1024, FortuneUsageText );
		return 1;
	}
	else
		return 0;
}

